/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/. */
"use strict";

const {makeObserver} = require("preferences");

exports.available = false;
let service = null;
let alerting = false;
let supportsClickable = true;
(function() {
	// XXX enhance: query libnotify via ctypes.
	if (require("version").OS == 'linux') {
		supportsClickable = false;
	}
})();

const Observer = {
	_func: null,
	observe: function(aSubject, aTopic, aData) {
		switch (aTopic) {
		case "alertfinished":
			alerting = false;
			break;
		case "alertclickcallback":
			if (this._func) {
				try {
					this._func();
				}
				catch (ex) {
					Cu.reportError(ex);
					// no op
				}
			}
			this._func = null;
			break;
		}
	}
};

try {
	service = Cc['@mozilla.org/alerts-service;1'].getService(Ci.nsIAlertsService);
	makeObserver(Observer);
	exports.available = true;
}
catch (ex) {
	log(LOG_ERROR, "alertsserivce", ex);
	// no-op
}

exports.show = function alertservice_show(title, msg, callback) {
	if (!exports.available) {
		throw new Exception("Alerting Service not available on this platform!");
	}
	if (alerting) {
		return;
	}

	let clickable = false;
	Observer._func = null;

	// don't make clickable on *nix, so that libnotify will be used more often
	if (typeof callback == 'function' && supportsClickable) {
		clickable = true;
		Observer._func = callback;
	}
	try {
		service.showAlertNotification(
			"chrome://dta/skin/common/alert.png",
			title,
			msg,
			clickable,
			null,
			Observer
			);
		alerting = true;
	}
	catch (ex if ex.result == Cr.NS_ERROR_NOT_IMPLEMENTED) {
		log(LOG_DEBUG, "alertsservice not available after all", ex);
		exports.available = false;
	}
	catch (ex if ex.result == Cr.NS_ERROR_NOT_AVAILABLE) {
		log(LOG_DEBUG, "alertsservice (temporarily) not available", ex);
	}
	catch (ex) {
		log(LOG_ERROR, "alertsservice unexpectedly failed", ex);
	}
}
